/*
 * Pound - the reverse-proxy load-balancer
 * Copyright (C) 2002 Apsis GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 * Contact information:
 * Apsis GmbH
 * P.O.Box
 * 8707 Uetikon am See
 * Switzerland
 * Tel: +41-1-920 4904
 * EMail: roseg@apsis.ch
 */

/*
 * $Id: pound.h,v 0.9 2002/08/19 08:19:53 roseg Exp roseg $
 *
 * $Log: pound.h,v $
 * Revision 0.9  2002/08/19 08:19:53  roseg
 * Added support for listening on multiple addresses/ports
 * Added support/configuration for WebDAV (LOCK/UNLOCK)
 * Added support for old-style HTTP/1.0 responses (content to EOF)
 * Fixed threads stack size problem on *BSD (#ifdef NEED_STACK)
 * Fixed problem in URL extraction
 *
 * Revision 0.8  2002/08/01 13:29:15  roseg
 * fixed bug in server timeout/close detection
 * fixed problem with SSL multi-threading
 * header collection
 * extended request patterns as per RFC
 * fixed problem with HEAD response (ignore content length)
 *
 * Revision 0.7  2002/07/23 03:11:28  roseg
 * Moved entirely to BIO (rather then the old comm_)
 * Added HTTPS-specific headers
 * Fixed a few minor problems in the pattern matching
 *
 * Revision 0.6  2002/07/16 21:14:01  roseg
 * added URL groups and matching
 * extended URL reuest matching
 * moved to "modern" regex
 *
 * Revision 0.5  2002/07/04 12:23:42  roseg
 * code split
 *
 */

#include    <stdio.h>
#include    <stdlib.h>
#include    <unistd.h>
#include    <pthread.h>
#include    <string.h>
#include    <sys/time.h>
#include    <sys/types.h>
#include    <time.h>
#include    <sys/socket.h>
#include    <netinet/in.h>
#include    <arpa/inet.h>
#include    <netdb.h>
#include    <openssl/ssl.h>
#include    <pwd.h>
#include    <grp.h>
#include    <syslog.h>
#include    <signal.h>
#include    <regex.h>
#include    <errno.h>

extern int errno;

/*
 * Global variables needed by everybody
 */

extern int  clnt_to;            /* client timeout */
extern int  log_level;          /* logging mode - 0, 1, 2 */
extern int  https_headers;      /* add HTTPS-specific headers */
extern int  allow_xtd;          /* allow extended HTTP - PUT, DELETE */
extern int  allow_dav;          /* allow WebDAV - LOCK, UNLOCK */
extern int  alive_to;           /* check interval for resurrection */
extern char **http,             /* HTTP port to listen on */
            **https,            /* HTTPS port to listen on */
            **cert,             /* certificate file */
            *user,              /* user to run as */
            *group,             /* group to run as */
            *root;              /* directory to chroot to */

#define MAXBUF      16378
#define MAXHEADERS  256
#define GLOB_SESS   15

/* Backend definition */
typedef struct {
    struct sockaddr_in  addr;   /* address */
    int                 alive;  /* alive check interval */
}   BACKEND;

/* Session definition */
typedef struct _sess {
    char    from_host[16];      /* client address */
    int     to_host;            /* backend index */
    time_t  last_acc;           /* time of last access */
    int     children;           /* number of children */
    struct _sess    *left, *right;
}   SESS;

#define n_children(S)   ((S)? (S)->children: 0)

/* URL group definition */
typedef struct _group {
    regex_t         url_pat;            /* pattern to match the URL against */
    BACKEND         *backend_addr;      /* array of backend servers */
    int             tot_pri;            /* total number of backend servers */
    int             sess_to;            /* session timeout */
    pthread_mutex_t mut;                /* group mutex */
    SESS            *sessions;          /* session tree root */
}   GROUP;

extern GROUP    **groups;

typedef struct  {
    int             sock;
    struct in_addr  from_host;
    int             is_ssl;
    char            *cert;
}   thr_arg;                        /* argument to processing threads: socket, origin */

extern regex_t  GET,                /* GET request: GET uri HTTP/1.[01] */
                POST,               /* POST request: POST uri HTTP/1.[01] */
                HEAD,               /* HEAD request: HEAD uri HTTP/1.[01] */
                PUT,                /* PUT request: PUT uri HTTP/1.[01] */
                DELETE,             /* DELETE request: DELETE uri HTTP/1.[01] */
                LOCK,               /* LOCK request: LOCK uri HTTP/1.[01] */
                UNLOCK,             /* UNLOCK request: UNLOCK uri HTTP/1.[01] */
                HEADER,             /* Allowed header */
                CHUNKED,            /* Transfer-encoding: chunked header */
                CONT_LEN,           /* Content-length header */
                CHUNK_HEAD;         /* chunk header line */

/*
 * handle an HTTP request
 */
extern void *thr_http(void *);

/*
 * Find the host to connect to
 */
extern struct sockaddr_in *get_be(struct in_addr, char *req_url);

/*
 * mark a backend host as dead;
 * do nothing if no resurection code is active
 */
extern void kill_be(struct sockaddr_in *);

/*
 * Prune the expired sessions and dead hosts from the table;
 * runs every session_to seconds (if needed)
 */
extern void *thr_prune(void *);

/*
 * Check if dead hosts returned to life;
 * runs every alive_to seconds
 */
extern void *thr_resurect(void *);

/*
 * Parse arguments/config file
 */
extern void config_parse(int, char **);
